/*
 * This file is part of decoding
 * Copyright 2012  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/* CARACTERISTIC p */

void ll_dbpol_shift_fast_rec(R *r,R *a,int nax,int nay,R y,R **pow,R *tmp) {
  int log2,nay0,nay1,offset;
  if ( nay <= 5 ) {
    R zero;
    R_init(zero);
    ll_dbpol_shift_naive(r,a,nax,nay,zero,y);
    R_clear(zero);
    return;
  }
  log2 = half_log2(nay);
  nay0 = 1 << log2;
  nay1 = nay - nay0;
  offset = nay0 * nax;
  ll_dbpol_shift_fast_rec(r,a,nax,nay0,y,pow,tmp);
  ll_dbpol_shift_fast_rec(&r[offset],&a[offset],nax,nay1,y,pow,tmp);
  ll_dbpol_mul_by_upolY(tmp,&r[offset],nax,nay1,pow[log2],nay0 + 1);
  ll_array_add(r,r,nay0 * nax,tmp,nay * nax);
}

/**
> `void ll_dbpol_shift_fast2(R *r,R *a,int nax,int nay,R x,R y,\
  struct ll_dbpol_shift_fast_pc *pc);`
  Set `r` to `a(X + x,Y + y)` where `a` has size (`nax`,`nay`)
  and `pc` is a pointer to a C structure where precomputations
  for the algorithm have been made by `ll_dbpol_shift_fast_prepare`.
  The memory occupied by `pc` is freed with
  `ll_dbpol_shift_fast_finish`.
  For example:

  ``
  struct ll_dbpol_shift_fast_pc *pc;
  pc = ll_dbpol_shift_fast_prepare(nax,nay,x,y);
  [ ... ]
  ll_dbpol_shift_fast2(r,a,nax,nay,x,y,pc);
  [ ... ]
  ll_dbpol_shift_fast_finish(pc);
  ``
*/

void ll_dbpol_shift_fast2(R *r,R *a,int nax,int nay,R x,R y,
                          struct ll_dbpol_shift_fast_pc *pc)
{
  int i;
  R *tmp,*atmp;
  /* shift in X in a temp buffer */
  atmp = ll_mat_malloc_init(nax,nay);
  for( i = 0 ; i < nay ; i++ ) {
    ll_upol_shift_fast2(&atmp[i * nax],&a[i * nax],nax,x,pc->powX);
  }
  /* go for it */
  tmp = ll_mat_malloc_init(nax,nay);
  /*
  ll_dbpol_shift_fast_rec(r,atmp,nax,nay,y,pc->powY->pow,tmp);
  */
  ll_mat_transpose(tmp,atmp,nay,nax);
  for( i = 0 ; i < nax ; i++ ) {
    ll_upol_shift_fast2(&atmp[i * nay],&tmp[i * nay],nay,y,pc->powY);
  }
  ll_mat_transpose(r,atmp,nax,nay);
  /* free stuff */
  ll_mat_free(tmp,nax,nay);
  ll_mat_free(atmp,nax,nay);
}

/**
> `struct ll_dbpol_shift_fast_pc \
  *ll_dbpol_shift_fast_prepare(int nax,int nay,R x,R y);`
  Return a pointer to a structure containing precomputations
  for the shift of dense bivariate polynomials of size (`nax`,`nay`)
  at (`x`,`y`).
*/

struct ll_dbpol_shift_fast_pc
*ll_dbpol_shift_fast_prepare(int nax,int nay,R x,R y) {
  struct ll_dbpol_shift_fast_pc *pc;  
  pc = malloc(sizeof(struct ll_dbpol_shift_fast_pc));
  pc->powX = ll_upol_shift_fast_prepare(nax,x);
  pc->powY = ll_upol_shift_fast_prepare(nay,y);
  return pc;
}

/**
> `void ll_dbpol_shift_fast_finish(struct ll_dbpol_shift_fast_pc *pc);`
  Free the memory occupied by `pc` where `pc` was returned by a call
  to `ll_dbpol_shift_fast_prepare`.
*/

void ll_dbpol_shift_fast_finish(struct ll_dbpol_shift_fast_pc *pc) {
  ll_upol_shift_fast_finish(pc->powX);
  ll_upol_shift_fast_finish(pc->powY);
  ll_free(pc);
}

/**
> `void ll_dbpol_shift_fast(R *r,R *a,int nax,int nay,R x,R y);`
  Set `r` to `a(X + x,Y + y)` where `a` has size (`nax`,`nay`).
  The dichotomy algorithm is used. Note that if your ring has
  caracteristic 2 prefer `ll_dbpol_shift_car22`.
*/

void ll_dbpol_shift_fast(R *r,R *a,int nax,int nay,R x,R y) {
  struct ll_dbpol_shift_fast_pc *pc;
  pc = ll_dbpol_shift_fast_prepare(nax,nay,x,y);
  ll_dbpol_shift_fast2(r,a,nax,nay,x,y,pc);
  ll_dbpol_shift_fast_finish(pc);
}

/* CARACTERISTIC 2 */

void ll_dbpol_shift_car2_rec(R *r,int nrx,int nry,R y,R *powY) {
  int log2,n,nry1,tmp;

  if ( nry <= 2 ) {
    switch(nry) {
      case 0:
      case 1:
        return;
      case 2:
        ll_vec_add_mul_by_cte(r,&r[nrx],nrx,y);
        return;
    }
  }
  log2 = half_log2(nry);
  n = 1 << log2;
  nry1 = nry - n;
  tmp = n * nrx;
  ll_dbpol_shift_car2_rec(r,nrx,n,y,powY);
  ll_dbpol_shift_car2_rec(&r[tmp],nrx,nry1,y,powY);
  ll_vec_add_mul_by_cte(r,&r[tmp],nrx * nry1,powY[log2]);
}

/**
> `void ll_dbpol_shift_car22(R *r,R *a,int nax,int nay,R x,R y,\
  struct ll_dbpol_shift_car2_pc *pc);`
  Set `r` to `a(X + x,Y + y)` where `a` has size (`nax`,`nay`)
  and `pc` is a pointer to a C structure where precomputations
  for the algorithm have been made by `ll_dbpol_shift_car2_prepare`.
  The memory occupied by `pc` is freed with
  `ll_dbpol_shift_car2_finish`.
  Note that this algorithm is specific for rings of caracterstic 2.
  A call to `ll_dbpol_shift_car22` with a ring of caracteristic not 2
  will return a *mathematically false* result.
  The *dichotomy* algorithm is used.
  For example:

  ``
  struct ll_dbpol_shift_car2_pc *pc;
  pc = ll_dbpol_shift_car2_prepare(nax,nay,x,y);
  [ ... ]
  ll_dbpol_shift_car22(r,a,nax,nay,x,y,pc);
  [ ... ]
  ll_dbpol_shift_car2_finish(pc);
  ``
*/

void ll_dbpol_shift_car22(R *r,R *a,int nax,int nay,R x,R y,
                          struct ll_dbpol_shift_car2_pc *pc)
{
  int i;
  /* shift in X first */
  for( i = 0 ; i < nay ; i++ ) {
    ll_upol_shift_car22(&r[i * nax],&a[i * nax],nax,x,pc->powX);
  }
  /* go for it */
  ll_dbpol_shift_car2_rec(r,nax,nay,y,pc->powY->pow);
}

/**
> `struct ll_dbpol_shift_car2_pc \
  *ll_dbpol_shift_car2_prepare(int nax,int nay,R x,R y);`
  Return a structure containing precomputations for the shift
  of dense bivariate polynomials of size (`nax`,`nay`) at
  (`x`,`y`). 
  Note that this function is specific for caracteristic 2.
*/

struct ll_dbpol_shift_car2_pc
*ll_dbpol_shift_car2_prepare(int nax,int nay,R x,R y) {
  struct ll_dbpol_shift_car2_pc *pc;
  pc = ll_malloc(sizeof(struct ll_dbpol_shift_car2_pc));
  pc->powX = ll_upol_shift_car2_prepare(nax,x);
  pc->powY = ll_upol_shift_car2_prepare(nay,y);
  return pc;
}

/**
> `void ll_dbpol_shift_car2(R *r,R *a,int nax,int nay,R x,R y);`
  Set `r` to `a(X + x,Y + y)` where `a` has size (`nax`,`nay`).
  Note that this function is specific for
  caracteristic 2.
  A call to `ll_dbpol_shift_car2` with a ring of caracteristic not 2
  will return a *mathematically false* result.
  The *dichotomy* algorithm is used.
*/

void ll_dbpol_shift_car2(R *r,R *a,int nax,int nay,R x,R y) {
  struct ll_dbpol_shift_car2_pc *pc;
  pc = ll_dbpol_shift_car2_prepare(nax,nay,x,y);
  ll_dbpol_shift_car22(r,a,nax,nay,x,y,pc);
  ll_dbpol_shift_car2_finish(pc);
}

/**
> `void ll_dbpol_shift_car2_finish(struct ll_dbpol_shift_car2_pc *pc)`
  Free the memory occupied by `pc` where `pc` was returned by a call
  to `ll_dbpol_shift_fast_prepare`.
  Note that this function is specific for caracteristic 2.
*/

void ll_dbpol_shift_car2_finish(struct ll_dbpol_shift_car2_pc *pc) {
  ll_upol_shift_car2_finish(pc->powX);
  ll_upol_shift_car2_finish(pc->powY);
  ll_free(pc);
}

