/*
 * This file is part of decoding
 * Copyright 2012  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/* Koetter interpolation algorithm. (Sudan variant)
   From: The Guruswami-Sudan Decoding Algorithm for Reed-Solomon Codes.
         R. J. McEliece. */

int ll_koetter_sudan(R *q,int *nqx,int *nqy,
                     R *supp,int n,int k,
                     int L,R *y,int tau)
{
  R **g,*f,delta,alphai,*bufR,*t,coeff,*xchg,*disc;
  int *wdeg,*ngy,*mark,*bufint;
  int ngx,ngy_max,i,j,j_star,ngy_star,wdeg_star,L1;
  int wdeg_max,ret;
  int ng_max,nbufR,pc_ngy;

  /* to make gcc happy */
  f = NULL;

  /* default ret code */
  ret = 0;
  
  /* malloc int's */
  L1 = L + 1;
  bufint = ll_malloc(3 * L1 * sizeof(int));
  wdeg = bufint;
  ngy = &bufint[L1];
  mark = &bufint[2 * L1];
  
  /* malloc R's */
  ngx = L1;
  ngy_max = (n - tau) + 1;
  ng_max = ngx * ngy_max;
  nbufR = (L1 + 1) * ng_max + L1;
  bufR = ll_vec_malloc_init(nbufR);
  t = &bufR[L1 * ng_max + L1];
  g = ll_malloc(L1 * sizeof(R*));
  disc = bufR;
  
  /* init other stuff */
  for( j = 0 ; j <= L ; j++ ) {
    g[j] = &bufR[j * ng_max + L1];
    mark[j] = 0;
    ngy[j] = 1;
    wdeg[j] = (k - 1) * j;
    R_castint(g[j][j],1);
  }
  wdeg_max = (ngy_max - 1) + (k - 1) * (ngx - 1);
  pc_ngy = 0;

  /* go for it */
  R_init(delta);
  R_init(coeff);
  R_init(alphai);
  for( i = 0 ; i < n ; i++ ) {

    /* evaluation to obtain the discrepancy */
    for( j = 0 ; j < L1 ; j++ ) {
      ll_dbpol_horner(disc[j],g[j],ngx,ngy[j],y[i],supp[i]);
    }

    /* main loop */

    /* find the smallest of the gj's */
    j_star = -1;
    ngy_star = ngy_max + 1;
    wdeg_star = wdeg_max + 1;
    for( j = 0 ; j <= L ; j++ ) {
      if ( mark[j] < 2 ) {
        if ( !R_iszero(disc[j]) ) {
          mark[j] = 1;
          if (   wdeg[j] <  wdeg_star ||
               ( wdeg[j] == wdeg_star && ngy[j] < ngy_star) )
          {
            f = g[j];
            j_star = j;
            ngy_star = ngy[j];
            wdeg_star = wdeg[j];
          }
        }
        else {
          mark[j] = 0;
        }
      }
    }

    /* update the gj's */
    if ( j_star > -1 ) {
      if ( R_inv(delta,disc[j_star]) ) {
        ll_die("Koetter(): delta is not invertible");
      }
      R_neg(delta,delta);
      for( j = 0 ; j <= L ; j++ ) {
        if ( mark[j] == 1 ) {
          if ( j == j_star ) {
            if ( ngy[j] >= ngy_max ) {
              mark[j] = 2;
            }
            else {
              R_neg(alphai,supp[i]);
              ll_dbpol_mul_by_affine(t,f,ngx,ngy_star,alphai);
            }
          }
          else {
            R_mul(coeff,delta,disc[j]);
            ll_array_add_mul_by_cte(g[j],ngx * ngy[j],
                                    f,ngx * ngy_star,coeff);
            ngy[j] = MAX(ngy[j],ngy_star);
          }
        }
      }
      xchg = t;
      t = g[j_star];
      g[j_star] = xchg;
      ngy[j_star]++;
      if ( ngy[j_star] > pc_ngy ) pc_ngy = ngy[j_star];
      wdeg[j_star]++;
    }
    else {
      ret = -1;
      goto ll_koetter_single_free_stuff;
    }
  }

  /* select the min of the gj's */
  j_star = -1;
  ngy_star = ngy_max + 1;
  wdeg_star = wdeg_max + 1;
  for( j = 0 ; j <= L ; j++ ) {
    if ( mark[j] < 2 ) {
      if ( wdeg[j] <  wdeg_star ||
         ( wdeg[j] == wdeg_star && ngy[j] < ngy_star) )
      {
        f = g[j];
        j_star = j;
        ngy_star = ngy[j];
        wdeg_star = wdeg[j];
      }
    }
  }
  if ( j_star == -1 ) {
    ret = -1;
    goto ll_koetter_single_free_stuff;
  }
  
  /* copy the candidate into q */
  ll_vec_copy(q,f,ngx * ngy_star);
  nqx[0] = ngx;
  nqy[0] = ngy_star;

ll_koetter_single_free_stuff:
  ll_free(bufint);
  ll_free(g);
  ll_vec_free(bufR,nbufR);
  R_clear(coeff);
  R_clear(delta);
  R_clear(alphai);
  
  return ret;
}

