/*
 * This file is part of decoding
 * Copyright 2012  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
> `void ll_print(FILE *out,const char *s,...);`
  Produce output into `out` according to the format `s`.
  The format is (for the moment) much simpler than `printf`'s.

  > `%e`
    Print an element of the ring.

  > `%v`
    Print a vector, its size must be passed as an argument.

  > `%m`
    Print a matrix, its size must be passed as an argument.

  > `%p`
    Print a univariate polynomial, its size must be passed as
    an argument.

  > `%q`
    Print a dense bivariate polynomial, its size must passed
    as an argument.

  > `%d`, `%c`, `%s`, `%f`
    Print an `int`, a `char`, a string (`char*`) and a `double`.

  For example:

  ``
  R *a;
  int na,nax,nay,ra,ca;
  int n;
  char b,*str;
  double d;
  
  ll_print(stdout,"A vector %v\\n",a,na);
  ll_print(stdout,"A matrix %m\\n",a,ra,ca);
  ll_print(stdout,"A univariate polynomial %p\\n",a,na);
  ll_print(stdout,"A dense bivariate polynomial %q\\n",a,nax,nay);
  ll_print(stdout,"And the usual integer %d,"
                  "char %c, string %s, double %f\\n",n,b,str,d);
  ``
*/

void ll_print(FILE *out,const char *s,...) {
  va_list ap;
  va_start(ap,s);
  for( ; s[0] ; s++ ) {
    if ( s[0] == '%' ) {
      s++;
      switch(s[0]) {
        case 0:
          fputc('%',out);
          va_end(ap);
          return;
        case '%':
          fputc('%',out);
          break;
        case ' ': {
          int a,i;
          a = va_arg(ap,int);
          for( i = 0 ; i < a ; i++ ) {
            fputc(' ',out);
          }
          break;
        }
        case 'c': {
          int a;
          a = va_arg(ap,int);
          fputc(a,out);
          break;
        }
        case 'd': {
          int a;
          a = va_arg(ap,int);
          fprintf(out,"%d",a);
          break;
        }
        case 'f': {
          double a;
          a = va_arg(ap,double);
          fprintf(out,"%f",a);
          break;
        }
        case 's': {
          char *a;
          a = va_arg(ap,char*);
          fprintf(out,"%s",a);
          break;
        }
        case 'e': {
          R_ptr a;
          a = va_arg(ap,R_ptr);
          R_print(out,a);
          break;
        }
        case 'p': {
          R *a;
          int na;
          a = va_arg(ap,R*);
          na = va_arg(ap,int);
          ll_upol_print(out,a,na);
          break;
        }
        case 'q' : {
          R *a;
          int nax,nay;
          a = va_arg(ap,R*);
          nax = va_arg(ap,int);
          nay = va_arg(ap,int);
          ll_dbpol_print_factorized(out,a,nax,nay);
          break;
        }
        case 'v': {
          R *a;
          int na;
          a = va_arg(ap,R*);
          na = va_arg(ap,int);
          ll_vec_print(out,a,na);
          break;
        }
        case 'm': {
          R *a;
          int ra,ca;
          a = va_arg(ap,R*);
          ra = va_arg(ap,int);
          ca = va_arg(ap,int);
          ll_mat_print(out,a,ra,ca);
          break;
        }
        default:
          break;
      }
    }
    else {
      fputc(s[0],out);
    }
  }
  va_end(ap);
}


