/*
 * This file is part of decoding
 * Copyright 2012  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/* CARACTERISTIC p */

void ll_upol_shift_fast_rec(R *r,R *a,int na,R b,R **pow,R *tmp) {
  int log2,n,na1;
  if ( na <= 5 ) {
    ll_upol_shift_naive(r,a,na,b);
    return;
  }
  log2 = half_log2(na);
  n = 1 << log2;
  na1 = na - n;
  ll_upol_shift_fast_rec(r,a,n,b,pow,tmp);
  ll_upol_shift_fast_rec(&r[n],&a[n],na1,b,pow,tmp);
  ll_upol_mul(tmp,&r[n],na1,pow[log2],n + 1);
  ll_array_add(r,r,n,tmp,na);
}

/**
> `void ll_upol_shift_fast2(R *r,R *a,int na,R b,\
  struct ll_upol_shift_fast_pc *pc);`
  Set `r` to `a(X + b)` where `a` has size `na`
  and `pc` is a pointer to a C structure where precomputations
  for the algorithm have been made by `ll_upol_shift_fast_prepare`.
  The memory occupied by `pc` is freed with
  `ll_upol_shift_fast_finish`.
  For example:

  ``
  struct ll_upol_shift_fast_pc *pc;;
  pc = ll_upol_shift_fast_prepare(na,x);
  [ ... ]
  ll_upol_shift_fast2(r,a,na,x,pc);
  [ ... ]
  ll_upol_shift_fast_finish(pc);
  ``
*/

void ll_upol_shift_fast2(R *r,R *a,int na,R b,
                         struct ll_upol_shift_fast_pc *pc)
{
  R *tmp;
  tmp = ll_vec_malloc_init(8 * na);
  ll_upol_shift_fast_rec(r,a,na,b,pc->pow,tmp);
  ll_vec_free(tmp,8 * na);
}

/**
> `struct ll_upol_shift_fast_pc \
  *ll_upol_shift_fast_prepare(int na,R x);`
  Return a pointer to a structure containing precomputations for
  the shift of univariate polynomials of size `na` at `x`.
*/

struct ll_upol_shift_fast_pc *ll_upol_shift_fast_prepare(int na,R x) {
  int npow,i;
  R **pow;
  struct ll_upol_shift_fast_pc *pc;

  npow = half_log2(na) + 1;
  if ( npow <= 0 ) npow = 1;
  pow = ll_malloc(npow * sizeof(R*));
  pow[0] = ll_vec_malloc_init(2);
  R_copy(pow[0][0],x);
  R_castint(pow[0][1],1);
  for( i = 1 ; i < npow ; i++ ) {
    pow[i] = ll_vec_malloc_init((1 << i) + 1);
    ll_upol_square(pow[i],pow[i - 1],(1 << (i - 1)) + 1);
  }
  pc = ll_malloc(sizeof(struct ll_upol_shift_fast_pc));
  pc->pow = pow;
  pc->npow = npow;
  return pc;
}

/**
> `void ll_upol_shift_fast_finish(struct ll_upol_shift_fast_pc *pc);`
  Free the memory occupied by `pc` where `pc` was returned by a call
  to `ll_upol_shift_fast_prepare`.
*/

void ll_upol_shift_fast_finish(struct ll_upol_shift_fast_pc *pc) {
  int i;
  for( i = 0 ; i < pc->npow ; i++ ) {
    ll_vec_free(pc->pow[i],(1 << (i + 1)));
  }
  ll_free(pc->pow);
  ll_free(pc);
}

/**
> `void ll_upol_shift_fast_finish(struct ll_upol_shift_fast_pc *pc);`
  Set `r` to `a(X + x)` where `a` has size `nax`.
  The dichotomy algorithm is used. Note that if your ring has
  caracteristic 2 prefer `ll_upol_shift_car22`.
*/

void ll_upol_shift_fast(R *r,R *a,int na,R x) {
  struct ll_upol_shift_fast_pc *pc;;
  pc = ll_upol_shift_fast_prepare(na,x);
  ll_upol_shift_fast2(r,a,na,x,pc);
  ll_upol_shift_fast_finish(pc);
}

/* CARACTERISTIC 2 */

void ll_upol_shift_car2_rec(R *r,R *a,int na,R b,R *pow) {
  int log2,n,na1;

  if ( na <= 2 ) {
    switch(na) {
      case 0:
        return;
      case 1:
        if ( a != r ) R_copy(r[0],a[0]);
        return;
      case 2:
        if ( a != r ) {
          R_copy(r[0],a[0]);
          R_copy(r[1],a[1]);
        }
        R_add_mul(r[0],a[1],b);
        return;
    }
  }
  log2 = half_log2(na);
  n = 1 << log2;
  na1 = na - n;
  ll_upol_shift_car2_rec(r,a,n,b,pow);
  ll_upol_shift_car2_rec(&r[n],&a[n],na1,b,pow);
  ll_vec_add_mul_by_cte(r,&r[n],na1,pow[log2]);
}

/**
> `void ll_upol_shift_car22(R *r,R *a,int na,R b,\
  struct ll_upol_shift_car2_pc *pc);`
  Set `r` to `a(X + b)` where `a` has size `na`
  and `pc` is a pointer to a C structure where precomputations
  for the algorithm have been made by `ll_upol_shift_car2_prepare`.
  The memory occupied by `pc` is freed with
  `ll_upol_shift_car2_finish`.
  Note that this algorithm is specific for rings of caracterstic 2.
  A call to `ll_upol_shift_car22` with a ring of caracteristic not 2
  will return a *mathematically false* result.
  The *dichotomy* algorithm is used.
  You can have `r == a`.
  For example:
  
  ``
  struct ll_upol_shift_car2_pc *pc;
  pc = ll_upol_shift_car2_prepare(na,x);
  [ ... ]
  ll_upol_shift_car22(r,a,na,x,pc);
  [ ... ]
  ll_upol_shift_car2_finish(pc);
  ``
*/

void ll_upol_shift_car22(R *r,R *a,int na,R b,
                         struct ll_upol_shift_car2_pc *pc)
{
  ll_upol_shift_car2_rec(r,a,na,b,pc->pow);
}

/**
> `struct ll_upol_shift_car2_pc \
  *ll_upol_shift_car2_prepare(int na,R x);`
  Return a structure containing precomputations for the shift
  of univariate polynomials of size `nax` at `x`. 
  Note that this function is specific for caracteristic 2.
*/

struct ll_upol_shift_car2_pc *ll_upol_shift_car2_prepare(int na,R x) {
  int npow,i;
  R *pow;
  struct ll_upol_shift_car2_pc *pc;

  npow = half_log2(na) + 1;
  if ( npow <= 0 ) npow = 1;
  pow = ll_vec_malloc_init(npow);
  R_copy(pow[0],x);
  for( i = 1 ; i < npow ; i++ ) {
    R_square(pow[i],pow[i - 1]);
  }
  pc = ll_malloc(sizeof(struct ll_upol_shift_car2_pc));
  pc->pow = pow;
  pc->npow = npow;
  return pc;
}

/**
> `void ll_upol_shift_car2_finish(struct ll_upol_shift_car2_pc *pc);`
  Free the memory occupied by `pc` where `pc` was returned by a call
  to `ll_upol_shift_fast_prepare`.
  Note that this function is specific for caracteristic 2.
*/

void ll_upol_shift_car2_finish(struct ll_upol_shift_car2_pc *pc) {
  ll_vec_free(pc->pow,pc->npow);
  ll_free(pc);
}

/**
> `void ll_upol_shift_car2(R *r,R *a,int na,R x);`
  Set `r` to `a(X + x)` where `a` has size `na`.
  Note that this function is specific for caracteristic 2.
  A call to `ll_upol_shift_car2` with a ring of caracteristic not 2
  will return a *mathematically false* result.
  The *dichotomy* algorithm is used.
*/

void ll_upol_shift_car2(R *r,R *a,int na,R x) {
  struct ll_upol_shift_car2_pc *pc;
  pc = ll_upol_shift_car2_prepare(na,x);
  ll_upol_shift_car22(r,a,na,x,pc);
  ll_upol_shift_car2_finish(pc);
}

