/*
 * This file is part of decoding
 * Copyright 2012  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
==OVERVIEW==

It is easy to control what algorithms are used by `decoding`, for
example if you want to replace unvariate polynomial multiplication
with your own function or if the ring you are implementing has
caracteristic 2.

==DESCRIPTION==

The file `include/decoding/rings/GF17.c` is an example of how to
use macros defined below.

> `#define R_NAME`
  `R` contains the name `NAME` of the ring you are implementing. See
  `decoding-rings`(3) for details.
*/

#ifdef R
#undef R
#endif

/**
> `#define RING_COMM_CAR2`
  Tell `decoding` that your ring has caracteristic 2. It is not
  mandatory to specify it. When defined it activates certain fast
  algorithms specific to characteristic 2.
*/

#ifdef RING_COMM_CAR2
#undef RING_COMM_CAR2
#endif

/**
> `#define RING_MODN_WORD`
  Tell `decoding` that the ring is `Z/nZ` where `n` holds
  within a machine word. It can correspond, for exemple, to the
  `unsigned char`, `unsigned short`, `unsigned int` or
  `unsigned long` C type. The `WORD_TYPE` macro must
  be defined to the C type that represents the elements of
  your ring.
*/

#ifdef RING_MODN_WORD
#undef RING_MODN_WORD
#ifdef WORD_TYPE
#undef WORD_TYPE
#endif
#endif

/**
> `#define CHARACTERISTIC_WORD`
  Tell `decoding` that the characteristic of the ring holds
  within a machine word.
*/

#ifdef CHARACTERISTIC_WORD
#undef CHARACTERISTIC_WORD
#endif

/**
> `#define CHARACTERISTIC_MPZ`
  Tell `decoding` that the characteristic of the ring is stored
  within a GMP multiprecision integer.
*/

#ifdef CHARACTERISTIC_MPZ
#undef CHARACTERISTIC_MPZ
#endif

/**
> `#define MY_UPOL_MUL myfunc`
  Replace default univariate polynomial multiplication by `myfunc`.
  See `decoding-lowlevels`(3) for the prototype of `myfunc`.
*/

#ifdef MY_UPOL_MUL
#undef MY_UPOL_MUL
#endif

/**
> `#define MY_UPOL_SQUARE myfunc`
  Replace default univariate polynomial squaring by `myfunc`.
  See `decoding-lowlevels`(3) for the prototype of `myfunc`.
*/

#ifdef MY_UPOL_SQUARE
#undef MY_UPOL_SQUARE
#endif

/**
> `#define MY_UPOL_SHIFT myfunc`
  Replace default univariate polynomial shifting by `myfunc`.
  See `decoding-lowlevels`(3) for the prototype of `myfunc`.
*/

#ifdef MY_UPOL_SHIFT
#undef MY_UPOL_SHIFT
#endif

/**
> `#define MY_DBPOL_MUL myfunc`
  Replace default dense bivariate polynomial multiplication by `myfunc`.
  See `decoding-lowlevels`(3) for the prototype of `myfunc`.
*/

#ifdef MY_DBPOL_MUL
#undef MY_DBPOL_MUL
#endif

/**
> `#define MY_DBPOL_SHIFT myfunc`
  Replace default dense bivariate polynomial shfting by `myfunc`.
  See `decoding-lowlevels`(3) for the prototype of `myfunc`.
*/

#ifdef MY_DBPOL_SHIFT
#undef MY_DBPOL_SHIFT
#endif

/**
> `#define MY_UPOL_ROOTS myfunc`
  Replace default univariate root finding by `myfunc`.
  See `decoding-lowlevels`(3) for the prototype of `myfunc`.
  Note that `decoding` provides no algorithm for univariate root
  finding.
*/

#ifdef MY_UPOL_ROOTS
#undef MY_UPOL_ROOTS
#endif

/**
> `#define MY_VEC_ADD myfunc`
  Replace default vector addition by `myfunc`.
  See `decoding-lowlevels`(3) for the prototype of `myfunc`.
*/

#ifdef MY_VEC_ADD
#undef MY_VEC_ADD
#endif

/**
> `#define MY_VEC_NEG myfunc`
  Replace default vector additive inverse computation by `myfunc`.
  See `decoding-lowlevels`(3) for the prototype of `myfunc`.
*/

#ifdef MY_VEC_NEG
#undef MY_VEC_NEG
#endif

/**
> `#define MY_VEC_SUB myfunc`
  Replace default vector substraction by `myfunc`.
  See `decoding-lowlevels`(3) for the prototype of `myfunc`.
*/

#ifdef MY_VEC_SUB
#undef MY_VEC_SUB
#endif

/**
> `#define MY_VEC_SUB_SUB myfunc`
  Replace default `ll_vec_sub_sub` by `myfunc`.
  See `decoding-lowlevels`(3) for the prototype of `myfunc`.
*/

#ifdef MY_VEC_SUB_SUB
#undef MY_VEC_SUB_SUB
#endif

/**
> `#define MY_VEC_ADD_MUL myfunc`
  Replace default `ll_vec_add_mul` by `myfunc`.
  See `decoding-lowlevels`(3) for the prototype of `myfunc`.
*/

#ifdef MY_VEC_ADD_MUL
#undef MY_VEC_ADD_MUL
#endif

/**
> `#define MY_VEC_MUL_BY_CTE myfunc`
  Replace default `ll_vec_mul_by_cte` by `myfunc`.
  See `decoding-lowlevels`(3) for the prototype of `myfunc`.
*/

#ifdef MY_VEC_MUL_BY_CTE
#undef MY_VEC_MUL_BY_CTE
#endif

/**
> `#define MY_VEC_ADD_MUL_BY_CTE myfunc`
  Replace default `ll_vec_add_mul_by_cte` by `myfunc`.
  See `decoding-lowlevels`(3) for the prototype of `myfunc`.
*/

#ifdef MY_VEC_ADD_MUL_BY_CTE
#undef MY_VEC_ADD_MUL_BY_CTE
#endif

/**
> `#define MY_VEC_SUB_MUL myfunc`
  Replace default `ll_vec_sub_mul` by `myfunc`.
  See `decoding-lowlevels`(3) for the prototype of `myfunc`.
*/

#ifdef MY_VEC_SUB_MUL
#undef MY_VEC_SUB_MUL
#endif

/**
===MEMORY MANAGEMENT===

`decoding` provides no garbage collectors by default.
The memory is manually allocated and freed with the
standard libc functions `stdlib.h`(0P). You can replace
the memory management functions with your own or
use an external garbage collector if you wish with the
`MY_*ALLOC` macros.

> `#define MY_MALLOC myfunc`
  Replace the `malloc` function by `myfunc`. By default
  the libc `malloc` is used. `myfunc` must have the same
  prototype as `malloc`.
  See `malloc`(3) for more details.

> `#define MY_REALLOC myfunc`
  Replace the `realloc` function by `myfunc`. By default
  the libc `realloc` is used. `myfunc` must have the same
  prototype as `realloc`.
  See `realloc`(3) for more details.

> `#define MY_CALLOC myfunc`
  Replace the `calloc` function by `myfunc`. By default
  the libc `calloc` is used. `myfunc` must have the same
  prototype as `calloc`.
  See `calloc`(3) for more details.

> `#define MY_FREE myfunc`
  Replace the `free` function by `myfunc`. By default
  the libc `free` is used. `myfunc` must have the same
  prototype as `free`.
  See `free`(3) for more details.

==AUTHOR==

Written by Guillaume Quintin (coincoin169g@gmail.com).

==SEE ALSO==

`decoding-rings`(3),`decoding-lowlevels`(3),
MAGMA(http://magma.maths.usyd.edu.au/magma/),
`malloc`(3), `realloc`(3), `calloc`(3), `free`(3),
`stdlib.h`(0P).
*/

