/*
 * This file is part of decoding
 * Copyright 2012  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
> `gfp_word_dynamic.c`
  Implement all prime fields (`Z/pZ`) whose characteristic holds
  within a machine word. The difference with `gfp_word.c` is that
  the prime defining the field can be chosen and modified at run
  time.

  ``
  #include <decoding/rings/gfp_word_dynamic.c>
  #include <decoding/algos.c>
  ``
  
  In the example above, the name of the finite field will be
  `gfp`. You can of course use several prime fields in the
  same program as soon as they have different names.

  ``
  #define RING_NAME fld1
  #include <decoding/rings/gfp_word_dynamic.c>
  #include <decoding/algos.c>
  
  #define RING_NAME fld2
  #include <decoding/rings/gfp_word_dynamic.c>
  #include <decoding/algos.c>
  ``

  The prime number defining the field is accessible with
  the `R_characteristic` variable. For exemple we will have access
  to the following variables:

  - `fld1_characteristic`.

  - `fdl2_characteristic`.

  By default, the `unsigned int` C type is choosen to hold the
  elements of the field. You can specify a different genuine
  integer C type like `unsigned short` or `unsigned long` if
  you need with `WORD_TYPE`. You *must* choose an unsigned type.

  ``
  #define WORD_TYPE unsigned long
  #include <decoding/rings/gfp_word_dynamic.c>
  #include <decoding/algos.c>
  ``
*/

#include <decoding/ring_reset.h>

/* standard stuff for rings */
#ifdef RING_NAME
#define R RING_NAME
#else
#define R gfp
#endif
#define RING_MODN_WORD

/* C genuine type choice, default to unsigned int */
#ifndef WORD_TYPE
#define WORD_TYPE unsigned int
#endif
#define CHARACTERISTIC_WORD

typedef WORD_TYPE *R_ptr;
typedef WORD_TYPE R[1];
typedef WORD_TYPE R_content;
typedef WORD_TYPE R_characteristic_type;
const char *R_magma_decl = __S(R) " := GF(" __S(R_characteristic) ")";
WORD_TYPE R_characteristic = 0;

/*========================================================*/
/* Vectorization stuff: for now we let gcc vectorize      */
/*========================================================*/
#define R_reduce_2p __C(R,_reduce_2p)
#define R_reduce_p2 __C(R,_reduce_p2)

#ifdef DECODING_OPTIMIZE
#define R_reduce_r __C(R,_reduce_r)
#define R_reduce_s __C(R,_reduce_s)
#define R_reduce_t __C(R,_reduce_t)
#define R_reduce_q __C(R,_reduce_q)
int R_reduce_r;
int R_reduce_s;
int R_reduce_t;
WORD_TYPE R_reduce_q;
#endif

WORD_TYPE R_reduce_2p(WORD_TYPE a) {
#ifdef DECODING_OPTIMIZE
  WORD_TYPE mask;
  mask = -(((R_characteristic - 1) - a) >> (sizeof(WORD_TYPE) * 8 - 1));
  return a - (R_characteristic & mask);
#else
  return a % R_characteristic;
#endif
}

WORD_TYPE R_reduce_p2(WORD_TYPE a) {
#ifdef DECODING_OPTIMIZE
  WORD_TYPE a1,b,c;
  a1 = a >> R_reduce_s;
  b = (a1 * R_reduce_q) >> R_reduce_t;
  c = a - b * (WORD_TYPE)R_characteristic;
  return R_reduce_2p(c);
#else
  return a % R_characteristic;
#endif
}
/*========================================================*/



/**
  > `void R_ring_init(R_characteristic p);`
    Initialize the ring with prime `p`.
    A call to this function is mandatory before doing anything else.
*/

void R_ring_init(R_characteristic_type p) {
#ifdef DECODING_OPTIMIZE
  int r;
  R_characteristic = p;
  r = half_log2((unsigned long)p) + 1;
  R_reduce_r = r;
  R_reduce_s = r - 2;
  R_reduce_t = r + 3;
  R_reduce_q = (WORD_TYPE)((1 << (2 * r + 1)) / p);
#else
  R_characteristic = p;
#endif
}

/**
  > `void R_ring_clear(void);`
    Free the memory occupied by the ring.
    A call to this function is recommended when the ring is
    no more needed.

    You can change the prime of the ring with successive calls
    to `R_ring_clear` and `R_ring_init`.

    ``
    #include <decoding/decoding.h>
    #define RING_NAME fld
    #include <decoding/rings/gfp_word_dynamic.c>
    
    ...

    int main(void) {
      fld_init_ring(101);
      ...
      fld_clear_ring();
      fld_init_ring(67);
      ...
      fld_clear_ring();
      return 0;
    }
    ``
*/

void R_ring_clear(void) {
}

void R_clear(R a) {
  (void)a;
}

void R_clears(R a,...) {
  (void)a;
}

void R_init(R a) {
  a[0] = 0;
}


void R_inits(R a,...) {
  va_list ap;
  va_start(ap,a);
  
  a[0] = 0;
  for(;;) {
    WORD_TYPE *b;
    b = va_arg(ap,WORD_TYPE*);
    if ( !b ) return;
    b[0] = 0;
  }
}

void R_copy(R r,R a) {
  r[0] = a[0];
}

void R_swap(R r,R a) {
  R_content tmp;
  tmp = r[0];
  r[0] = a[0];
  a[0] = tmp;
}

void R_init_copy(R r,R a) {
  r[0] = a[0];
}

void R_shallow_copy(R r,R a) {
  r[0] = a[0];
}

void R_print(FILE *out,R a) {
  fprintf(out,"%d",a[0]);
}

int R_cmp(R a,R b) {
  return (a[0] != b[0]);
}

int R_iszero(R a) {
  return (a[0] == 0);
}

void R_zero(R r) {
  r[0] = 0;
}

void R_random(R r) {
  r[0] = rand() % R_characteristic;
}

void R_add(R r,R a,R b) {
  r[0] = R_reduce_2p(a[0] + b[0]);
}

void R_sub(R r,R a,R b) {
  r[0] = R_reduce_2p(a[0] + R_characteristic - b[0]);
}

void R_sub_sub(R r,R a,R b) {
  r[0] = R_reduce_2p(r[0] + R_characteristic - a[0]);
  r[0] = R_reduce_2p(r[0] + R_characteristic - b[0]);
}

void R_add_mul(R r,R a,R b) {
  r[0] = R_reduce_2p(r[0] + R_reduce_p2(a[0] * b[0]));
}

void R_sub_mul(R r,R a,R b) {
  r[0] = R_reduce_2p(r[0] + R_reduce_p2((R_characteristic * R_characteristic) - (a[0] * b[0])));
}

void R_neg(R r,R a) {
  r[0] = R_characteristic - a[0];
}

void R_mul(R r,R a,R b) {
  r[0] = R_reduce_p2(a[0] * b[0]);
}

void R_square(R r,R a) {
  r[0] = R_reduce_p2(a[0] * a[0]);
}

void R_castint(R r,int a) {
  if ( a >= 0 ) {
    r[0] = (WORD_TYPE)(a % R_characteristic);
  }
  else {
    r[0] = (WORD_TYPE)(R_characteristic - ((-a) % R_characteristic));
  }
}

void R_cast_integer(R r,R_characteristic_type a) {
  r[0] = (WORD_TYPE)(a % R_characteristic);
}

int R_inv(R r,R a) {
  WORD_TYPE q,r0,r1,r2,s0,s1,s2,t0,t1,t2;
  if ( a[0] == 0 ) return -1;
  r0 = R_characteristic; r1 = a[0];
  s0 = 1;     s1 = 0;
  t0 = 0;     t1 = 1;
  for( ; r1 != 0 ; ) {
    q = r0 / r1;
    r2 = r0 - q * r1;
    s2 = s0 - q * s1;
    t2 = t0 - q * t1;
    r0 = r1; r1 = r2;
    s0 = s1; s1 = s2;
    t0 = t1; t1 = t2;
  }
  R_castint(r,t0);
  return 0;
}

int R_div(R r,R a,R b) {
  R c;
  if ( R_inv(c,b) < 0 ) return -1;
  R_mul(r,a,c);
  return 0;
}

void R_next(R r,R a) {
  r[0] = a[0] + 1;
  if ( r[0] == R_characteristic ) r[0] = 0;
}

#include <decoding/lowlevels_decl.h>

#define MY_UPOL_ROOTS __C(R,_my_upol_roots)
void MY_UPOL_ROOTS(R *r,int *nr,R *a,int na) {
  R_content i;
  R x,y;
  nr[0] = 0;
  for( i = 0 ; i < R_characteristic ; i++ ) {
    x[0] = i;
    ll_upol_horner(y,a,na,x);
    if ( y[0] == 0 ) {
      R_copy(r[nr[0]],x);
      nr[0]++;
    }
  }
}

#undef R_reduce_2p
#undef R_reduce_p2

#ifdef DECODING_OPTIMIZE
#undef R_reduce_r
#undef R_reduce_s
#undef R_reduce_t
#undef R_reduce_q
#endif

