/*
 * This file is part of decoding
 * Copyright 2012  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <decoding/ring_reset.h>

#ifdef RING_NAME
#define R RING_NAME
#else
#define R gf2_50
#endif
#define RING_COMM_CAR2
#define CHARACTERISTIC_WORD

typedef mpfq_2_50_dst_elt R_ptr;
typedef mpfq_2_50_elt R;
typedef unsigned long R_content;
typedef int R_characteristic_type;
const int R_characteristic = 2;
const char *R_magma_decl = __S(R) " := GF(2,50)";

#define R_ext_degree __C(R,_ext_degree)
const int R_ext_degree = 50;

#define R_mpfq_field __C(R,_mpfq_field)
mpfq_2_50_field R_mpfq_field;

void R_ring_init(void) {
  mpfq_2_50_field_init(R_mpfq_field);
}

void R_ring_clear(void) {
  mpfq_2_50_field_clear(R_mpfq_field);
}

void R_clear(R_ptr a) {
  mpfq_2_50_clear(R_mpfq_field,&a);
}

void R_clears(R a,...) {
  va_list ap;
  va_start(ap,a);
  R_clear(a);
  for(;;) {
    R_ptr b;
    b = va_arg(ap,R_ptr);
    if ( !b ) return;
    R_clear(b);
  }
}

void R_init(R_ptr a) {
  mpfq_2_50_init(R_mpfq_field,&a);
  mpfq_2_50_set_ui(R_mpfq_field,a,0);
}

void R_inits(R a,...) {
  va_list ap;
  va_start(ap,a);
  R_init(a);
  for(;;) {
    R_ptr b;
    b = va_arg(ap,R_ptr);
    if ( !b ) return;
    R_init(b);
  }
}

void R_print(FILE *out,R a) {
  mpfq_2_50_fprint(R_mpfq_field,out,a);
}

void R_mul(R c,R a,R b) {
  mpfq_2_50_mul(R_mpfq_field,c,a,b);
}

int R_inv(R c,R a) {
  return !mpfq_2_50_inv(R_mpfq_field,c,a);
}

int R_div(R r,R a,R b) {
  R tmp;
  R_init(tmp);
  if ( R_inv(tmp,b) ) {
    R_clear(tmp);
    return -1;
  }
  R_mul(r,a,tmp);
  R_clear(tmp);
}

void R_copy(R r,R a) {
  mpfq_2_50_set(R_mpfq_field,r,a);
}

void R_swap(R r,R a) {
  R tmp;
  R_copy(tmp,r);
  R_copy(r,a);
  R_copy(a,tmp);
  R_clear(tmp);
}

void R_init_copy(R r,R a) {
  R_init(r);
  R_copy(r,a);
}

void R_shallow_copy(R r,R a) {
  R_copy(r,a);
}

int R_cmp(R a,R b) {
  return mpfq_2_50_cmp(R_mpfq_field,a,b);
}

int R_iszero(R a) {
  return !mpfq_2_50_cmp_ui(R_mpfq_field,a,0);
}

void R_zero(R r) {
  mpfq_2_50_set_ui(R_mpfq_field,r,0);
}

void R_random(R r) {
  mpfq_2_50_random(R_mpfq_field,r);
}

void R_add(R r,R a,R b) {
  mpfq_2_50_add(R_mpfq_field,r,a,b);
}

void R_sub(R r,R a,R b) {
  mpfq_2_50_add(R_mpfq_field,r,a,b); /* no an error, not proveided by mpfq */
}

void R_sub_sub(R r,R a,R b) {
  R tmp;
  R_init(tmp);
  R_sub(tmp,r,a);
  R_sub(r,tmp,b);
  R_clear(tmp);
}

void R_add_mul(R r,R a,R b) {
  R tmp;
  R_init(tmp);
  R_mul(tmp,a,b);
  R_add(r,r,tmp);
  R_clear(tmp);
}

void R_sub_mul(R r,R a,R b) {
  R tmp;
  R_init(tmp);
  R_mul(tmp,a,b);
  R_sub(r,r,tmp);
  R_clear(tmp);
}

void R_neg(R r,R a) {
  mpfq_2_50_neg(R_mpfq_field,r,a);
}

void R_square(R r,R a) {
   mpfq_2_50_sqr(R_mpfq_field,r,a);
}

void R_castint(R r,int a) {
  if ( a >= 0 ) {
    mpfq_2_50_set_ui(R_mpfq_field,r,(unsigned long)a);
  }
  else {
    mpfq_2_50_set_ui(R_mpfq_field,r,(unsigned long)(-a));
  }
}

void R_cast_integer(R r,int a) {
  R_castint(r,a);
}

#define SZ    (sizeof(unsigned long) * 8)
#define DEG   mpfq_2_50_field_degree(R_mpfq_field)
#define MAXUI ((unsigned long)-1)
#define MASK  (MAXUI >> (SZ - (DEG % SZ)))
#define LEN   ((DEG - 1) / sizeof(unsigned long) + 1)

void R_next(R r,R a) {
  int i;
  unsigned long carry;
  carry = 0;
  if ( a[0] == MAXUI ) carry = 1;
  r[0] = a[0] + 1;
  for( i = 1 ; i < LEN ; i++ ) {
    r[i] = a[i] + carry;
    if ( a[i] != MAXUI || carry == 0 ) carry = 0;
  }
  r[LEN - 1] &= MASK;
}

#undef LEN
#undef MASK
#undef MAXUI
#undef DEG
#undef SZ

#include <decoding/lowlevels_decl.h>

#define MY_UPOL_ROOTS __C(R,_my_upol_roots)
void MY_UPOL_ROOTS(R *r,int *nr,R *a,int na) {
  int i,n;
  R x,y;
  n = 0;
  R_init(x);
  R_init(y);
  for(;;) {
    ll_upol_horner(y,a,na,x);
    if ( R_iszero(y) ) {
      R_copy(r[n],x);
      n++;
    }
    R_next(x,x);
    if ( R_iszero(x) ) break;
  }
  R_clear(y);
  R_clear(x);
  nr[0] = n;
}

#undef R_mpfq_field
#undef R_ext_degree

