/*
 * This file is part of decoding
 * Copyright 2012  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
> `void rs_code_init(rs_code rs,int n,int k);`
  Initialize `rs` of length `n` and dimension `k`.
  This function must be called before doing anything else with `rs`.
  The support of `rs` will be a zero vector.
*/

void rs_code_init(rs_code rs,int n,int k) {
  int tau,m;
  
  rs[0].n = n;
  rs[0].k = k;
  rs[0].tmp = ll_vec_malloc_init(n);
  rs[0].ntmp = n;
  vec_init(rs[0].supp);
  vec_adjust_size(rs[0].supp,n);
  
  tau = sudan_max_errors(n,k);
  rs[0].stau = tau;
  rs[0].sL = sudan_list_size(n,k,tau);
  rs[0].snx = n - tau;

  tau = guruswami_sudan_max_errors(n,k);
  m = guruswami_sudan_mult(n,k,tau);
  rs[0].gstau = tau;
  rs[0].gsL = guruswami_sudan_list_size(n,k,tau,m);
  rs[0].gsm = m;
  rs[0].gsnx = m * (n - tau);
}

/**
> `void rs_code_init_with_support(rs_code rs,int n,int k);`
  Initialize `rs` of length `n` and dimension `k`.
  This function must be called before doing anything else with `rs`.
  `Decoding` will attempt to create the support of `rs`:
  it will first set the first coordinate to zero, then
  it will use `R_next` to set the other coordinates.
  Make sure that the ring you are using has enough elements.
  For exemple, if the ring is a prime field `gfp` then the
  support will be set to `[0,1,2,3,...,n]`.
*/

void rs_code_init_with_support(rs_code rs,int n,int k) {
  int i;
  R *supp;
  rs_code_init(rs,n,k);
  supp = rs[0].supp[0].a;
  R_zero(supp[0]);
  for( i = 1 ; i < n ; i++ ) {
    R_next(supp[i],supp[i - 1]);
  }
}

/**
> `void rs_code_clear(rs_code rs);`
  Free the memory occupied by `rs`.
*/

void rs_code_clear(rs_code rs) {
  vec_clear(rs[0].supp);
  /*
  ll_vec_free(rs[0].tmp,rs[0].ntmp);
  ll_vec_free(rs[0].ldec,rs[0].nldec);
  */
}

/**
> `void rs_code_random(vec r,rs_code rs);`
  Set `r` to be a random codeword of `rs`.
*/

void rs_code_random_codeword(vec r,rs_code rs) {
  R *tmp;
  int n,k;
  n = rs[0].n;
  k = rs[0].k;
  tmp = rs[0].tmp;
  ll_vec_random(tmp,k);
  vec_adjust_size(r,n);
  ll_upol_multi_horner(r[0].a,tmp,k,rs[0].supp[0].a,n);
}

/**
> `int rs_code_sudan_koetter(vec **r,rs_code rs,vec y,int tau);`
  Given a Reed-Solomon code `rs` and a received word `y` with at most
  `tau` errors, return the number of codewords within distance `tau`
  of `y` and set `r` to an array containing the codewords.
  The Sudan algorithm is used, the interpolation step is done with the
  Koetter algorithm.
*/

int rs_code_sudan_koetter(vec **r,rs_code rs,vec y,int tau) {
  int n,k,L,tau2;
  int nqx,nqy,nq;
  R *yptr,*q,*supp,*tmp;
  struct ll_lili_vec *roots,*v;
  vec *c;
  int nc;
  
  /* stuff with the parameters */
  n = rs[0].n;
  k = rs[0].k;
  yptr = y[0].a;
  supp = rs[0].supp[0].a;
  if ( tau == SUDAN_MAX_ERRORS ) {
    L = rs[0].sL;
    tau2 = rs[0].stau;
  }
  else {
    L = sudan_list_size(n,k,tau);
    tau2 = tau;
  }
  nq = (sudan_list_size(n,k,tau2) + 1) * (n - tau2 + 1);
  q = ll_vec_malloc_init(nq);
  
  /* 1) interpolation step */
  if ( ll_koetter_single(q,&nqx,&nqy,supp,n,k,L,1,yptr,tau2) ) {
    return 0;
  }
  
  /* 2) root finding */
  roots = ll_dbpol_Xroots(q,nqx,nqy,k);
  
  /* 3) select the roots within distance tau of y */
  if ( !roots ) return 0;
  nc = 0;
  tmp = 0;
  c = 0;
  for( v = roots ; v ; v = v->next ) {
    if ( !tmp ) tmp = ll_vec_malloc_init(n);
    ll_upol_multi_horner(tmp,v->a,k,supp,n);
    if ( ll_vec_hamming_distance(yptr,tmp,n) <= tau2 ) {
      if ( nc == 0 ) {
        c = ll_malloc(sizeof(vec));
      }
      else {
        c = ll_realloc(c,(nc + 1) * sizeof(vec));
      }
      vec_init(c[nc]);
      c[nc][0].a = tmp;
      c[nc][0].na = n;
      nc++;
      tmp = 0;
    }
  }
  *r = c;

  /* free stuff */
  ll_vec_free(q,nq);

  return nc;
}

/**
> `int rs_code_guruswami_sudan_koetter(vec **r,rs_code rs,vec y,int tau);`
  Given a Reed-Solomon code `rs` and a received word `y` with at most
  `tau` errors, return the number of codewords within distance `tau`
  of `y` and set `r` to an array containing the codewords.
  The Guruswami-Sudan algorithm is used, the interpolation step is done
  with the Koetter algorithm.
*/

int rs_code_guruswami_sudan_koetter(vec **r,rs_code rs,vec y,int tau) {
  int n,k,L,m,tau2;
  int nqx,nqy,nq;
  R *yptr,*q,*supp,*tmp;
  struct ll_lili_vec *roots,*v;
  vec *c;
  int nc;
  
  /* stuff with the parameters */
  n = rs[0].n;
  k = rs[0].k;
  yptr = y[0].a;
  supp = rs[0].supp[0].a;
  m = guruswami_sudan_mult(n,k,tau);
  if ( tau == GURUSWAMI_SUDAN_MAX_ERRORS ) {
    L = rs[0].gsL;
    tau2 = rs[0].gstau;
    m = rs[0].gsm;
  }
  else {
    m = guruswami_sudan_mult(n,k,tau);
    L = guruswami_sudan_list_size(n,k,tau,m);
    tau2 = tau;
  }
  nq = (guruswami_sudan_list_size(n,k,tau2,m) + 1) *
       (m * (n - tau2) + 1);
  q = ll_vec_malloc_init(nq);

  /* 1) interpolation step */
  if ( ll_koetter_single(q,&nqx,&nqy,supp,n,k,L,m,yptr,tau2) ) {
    return 0;
  }

  /* 2) root finding */
  roots = ll_dbpol_Xroots(q,nqx,nqy,k);

  /* 3) select the roots within distance tau of y */
  if ( !roots ) return 0;
  nc = 0;
  tmp = 0;
  c = 0;
  for( v = roots ; v ; v = v->next ) {
    if ( !tmp ) tmp = ll_vec_malloc_init(n);
    ll_upol_multi_horner(tmp,v->a,k,supp,n);
    if ( ll_vec_hamming_distance(yptr,tmp,n) <= tau2 ) {
      if ( nc == 0 ) {
        c = ll_malloc(sizeof(vec));
      }
      else {
        c = ll_realloc(c,(nc + 1) * sizeof(vec));
      }
      vec_init(c[nc]);
      c[nc][0].a = tmp;
      c[nc][0].na = n;
      nc++;
      tmp = 0;
    }
  }
  *r = c;

  /* free stuff */
  ll_vec_free(q,nq);

  return nc;
}

