/*
 * This file is part of decoding
 * Copyright 2012  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#define _POSIX_C_SOURCE 1

#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <decoding/decoding.h>

#define RING_NAME fld
#define PRIME 23
#include <decoding/rings/gfp_word.c>
#include <decoding/algos.c>

void mtest_dbpol_mul_naive(FILE *out,int n1,int n2) {
  int i;
  fld *f,*g,*fg;
  f = fld_ll_dbpol_malloc_random(n2,n2);
  g = fld_ll_dbpol_malloc_random(n2,n2);
  fg = fld_ll_mat_malloc_init(2 * n2 - 1,2 * n2 - 1);
  for( i = n1 ; i <= n2 ; i++ ) {
    fld_ll_dbpol_mul_naive(fg,f,i,i,g,i,i);
    fld_ll_print(out,"if e eq 0 then\n"
                     "  f  := P ! (%q);\n"
                     "  g  := P ! (%q);\n"
                     "  fg := P ! (%q);\n"
                     "  if not (fg eq (f * g)) then\n"
                     "    print \"Error\";\n"
                     "    e := 1;\n"
                     "  end if;\n"
                     "end if;\n\n",
                     f,i,i,g,i,i,fg,2 * i - 1,2 * i - 1);
  }
  fld_ll_vec_free(f,n2);
  fld_ll_vec_free(g,n2);
  fld_ll_mat_free(fg,2 * n2 - 1,2 * n2 - 1);
}

void mtest_dbpol_mul_by_affine(FILE *out,int n1,int n2) {
  int i;
  fld *f,*g,b;
  fld_init(b);
  f = fld_ll_dbpol_malloc_random(n2,n2);
  g = fld_ll_mat_malloc_init(n2,n2 + 1);
  for( i = n1 ; i <= n2 ; i++ ) {
    fld_random(b);
    fld_ll_dbpol_mul_by_affine(g,f,i,i,b);
    fld_ll_print(out,"if e eq 0 then\n"
                     "  f := P ! (%q);\n"
                     "  g := P ! (%q);\n"
                     "  b := fld ! (%e);\n"
                     "  if not (f * (Y + b) eq g) then\n"
                     "    print \"Error\";\n"
                     "    e := 1;\n"
                     "  end if;\n"
                     "end if;\n\n",
                     f,i,i,g,i,i + 1,b);
  }
  fld_ll_mat_free(f,n2,n2);
  fld_ll_mat_free(g,n2,n2 + 1);
  fld_clear(b);
}

void mtest_dbpol_mul_by_affine_inplace(FILE *out,int n1,int n2) {
  int i;
  fld  *f,*g,b;
  fld_init(b);
  f = fld_ll_dbpol_malloc_random(n2,n2);
  g = fld_ll_mat_malloc_init(n2,n2 + 1);
  for( i = n1 ; i <= n2 ; i++ ) {
    fld_random(b);
    fld_ll_mat_copy(g,f,i,i);
    fld_ll_dbpol_mul_by_affine(g,g,i,i,b);
    fld_ll_print(out,"if e eq 0 then\n"
                     "  f := P ! (%q);\n"
                     "  g := P ! (%q);\n"
                     "  b := fld ! (%e);\n"
                     "  if not (f * (Y + b) eq g) then\n"
                     "    print \"Error\";\n"
                     "    e := 1;\n"
                     "  end if;\n"
                     "end if;\n\n",
                     f,i,i,g,i,i + 1,b);
  }
  fld_ll_mat_free(f,n2,n2);
  fld_ll_mat_free(g,n2,n2 + 1);
  fld_clear(b);
}

void mtest_dbpol_shift_naive(FILE *out,int n1,int n2) {
  int i,j;
  fld  *f,*g,x,y;
  f = fld_ll_dbpol_malloc_random(n2,n2);
  g = fld_ll_mat_malloc_init(n2,n2);
  fld_init(x);
  fld_init(y);
  for( i = n1 ; i <= n2 ; i++ ) {
    fld_random(x);
    fld_random(y);
    j = i / 3;
    fld_ll_dbpol_shift_naive(g,f,i,j,x,y);
    fld_ll_print(out,"if e eq 0 then\n"
                     "  f := P ! (%q);\n"
                     "  x := fld ! (%e);\n"
                     "  y := fld ! (%e);\n"
                     "  g := P ! (%q);\n"
                     "  if not (Evaluate(f,[X + x,Y + y]) eq g) then\n"
                     "    print \"Error\";\n"
                     "    e := 1;\n"
                     "  end if;\n"
                     "end if;\n\n",
                     f,i,j,x,y,g,i,j);
  }
  fld_ll_mat_free(f,n2,n2);
  fld_ll_mat_free(g,n2,n2);
  fld_clear(y);
  fld_clear(x);
}

void mtest_dbpol_shift_fast(FILE *out,int n1,int n2) {
  int i;
  fld  *f,*g,x,y;
  f = fld_ll_dbpol_malloc_random(n2,n2);
  g = fld_ll_mat_malloc_init(n2,n2);
  fld_init(x);
  fld_init(y);
  for( i = n1 ; i <= n2 ; i++ ) {
    fld_random(x);
    fld_random(y);
    fld_ll_dbpol_shift_fast(g,f,i,i,x,y);
    fld_ll_print(out,"if e eq 0 then\n"
                     "  f := P ! (%q);\n"
                     "  x := fld ! (%e);\n"
                     "  y := fld ! (%e);\n"
                     "  g := P ! (%q);\n"
                     "  if not (Evaluate(f,[X + x,Y + y]) eq g) then\n"
                     "    print \"Error\";\n"
                     "    e := 1;\n"
                     "  end if;\n"
                     "end if;\n\n",
                     f,i,i,x,y,g,i,i);
  }
  fld_ll_mat_free(f,n2,n2);
  fld_ll_mat_free(g,n2,n2);
  fld_clear(y);
  fld_clear(x);
}

void mtest_dbpol_horner_upol(FILE *out,int n1,int n2) {
  int i,ng,nh;
  fld  *f,*g,*h;
  f = fld_ll_dbpol_malloc_random(n2,n2);
  g = fld_ll_vec_malloc_init(n2);
  h = fld_ll_vec_malloc_init(n2 + n2 * n2);
  for( i = n1 ; i <= n2 ; i++ ) {
    ng = MAX(2,i / 10);
    fld_ll_upol_random(g,ng);
    fld_ll_dbpol_horner_upol(h,&nh,f,i,i,g,ng);
    fld_ll_print(out,"if e eq 0 then\n"
                     "  f := P ! (%q);\n"
                     "  g := P ! (%p);\n"
                     "  h := P ! (%p);\n"
                     "  if not (Evaluate(f,[X,g]) eq h) then\n"
                     "    print \"Error\";\n"
                     "    e := 1;\n"
                     "  end if;\n"
                     "end if;\n\n",
                     f,i,i,g,ng,h,nh);
  }
  fld_ll_mat_free(f,n2,n2);
  fld_ll_vec_free(g,n2);
  fld_ll_vec_free(h,n2 + n2 * n2);
}

void mtest_dbpol_replace_X_by_XYn(FILE *out,int n1,int n2) {
  int i,n,ngy;
  fld  *f,*g;
  for( n = 0 ; n <= 10 ; n++ ) {
    f = fld_ll_dbpol_malloc_random(n2,n2);
    g = fld_ll_mat_malloc_init(n2,n2 + n * n2);
    for( i = n1 ; i <= n2 ; i++ ) {
      fld_ll_dbpol_replace_X_by_XYn(g,&ngy,f,i,i,n);
      fld_ll_print(out,"if e eq 0 then\n"
                       "  f := P ! (%q);\n"
                       "  g := P ! (%q);\n"
                       "  if not (Evaluate(f,[X * Y^%d,Y]) eq g) then\n"
                       "    print \"Error\";\n"
                       "    e := 1;\n"
                       "  end if;\n"
                       "end if;\n\n",
                       f,i,i,g,i,ngy,n);
    }
    fld_ll_mat_free(f,n2,n2);
    fld_ll_mat_free(g,n2,n2 + n * n2);
  }
}

void mtest_dbpol_replace_X_by_XYn_in_place(FILE *out,int n1,int n2) {
  int i,n,ngy;
  fld  *f,*g;
  for( n = 0 ; n <= 10 ; n++ ) {
    f = fld_ll_dbpol_malloc_random(n2,n2);
    g = fld_ll_mat_malloc_init(n2,n2 + n * n2);
    for( i = n1 ; i <= n2 ; i++ ) {
      fld_ll_mat_copy(g,f,i,i);
      fld_ll_dbpol_replace_X_by_XYn(g,&ngy,g,i,i,n);
      fld_ll_print(out,"if e eq 0 then\n"
                       "  f := P ! (%q);\n"
                       "  g := P ! (%q);\n"
                       "  if not (Evaluate(f,[X * Y^%d,Y]) eq g) then\n"
                       "    print \"Error\";\n"
                       "    e := 1;\n"
                       "  end if;\n"
                       "end if;\n\n",
                       f,i,i,g,i,ngy,n);
    }
    fld_ll_mat_free(f,n2,n2);
    fld_ll_mat_free(g,n2,n2 + n * n2);
  }
}

void mtest_dbpol_shift_in_X(FILE *out,int n1,int n2) {
  int i;
  fld  *f,*g,x;
  f = fld_ll_dbpol_malloc_random(n2,n2);
  g = fld_ll_mat_malloc_init(n2,n2);
  fld_init(x);
  for( i = n1 ; i <= n2 ; i++ ) {
    fld_random(x);
    fld_ll_dbpol_shift_in_X(g,f,i,i,x);
    fld_ll_print(out,"if e eq 0 then\n"
                     "  f := P ! (%q);\n"
                     "  x := fld ! (%e);\n"
                     "  g := P ! (%q);\n"
                     "  if not (Evaluate(f,[X + x,Y]) eq g) then\n"
                     "    print \"Error\";\n"
                     "    e := 1;\n"
                     "  end if;\n"
                     "end if;\n\n",
                     f,i,i,x,g,i,i);
  }
  fld_ll_mat_free(f,n2,n2);
  fld_ll_mat_free(g,n2,n2);
  fld_clear(x);
}


#define D1 3
#define D2 50

int main(void) {
  srand(time(0));
  fld_ring_init();
  fld_ll_print(stdout,"clear;\n"
                      "%s;\n"
                      "P<X,Y> := PolynomialRing(fld,2);\n"
                      "e := 0;\n\n",fld_magma_decl);
  /*
  mtest_dbpol_mul_naive(stdout,D1,D2);
  mtest_dbpol_mul_by_affine(stdout,D1,D2);
  mtest_dbpol_mul_by_affine_inplace(stdout,D1,D2);
  */
  mtest_dbpol_shift_naive(stdout,D1,D2);
  mtest_dbpol_shift_fast(stdout,D1,D2);
  /*
  mtest_dbpol_horner_upol(stdout,D1,D2);
  mtest_dbpol_replace_X_by_XYn(stdout,D1,D2);
  mtest_dbpol_replace_X_by_XYn_in_place(stdout,D1,D2);
  mtest_dbpol_shift_in_X(stdout,D1,D2);
  */
  fld_ll_print(stdout,"if e eq 0 then\n"
                      "  print \"ALL OK\";\n"
                      "end if;\n\n");
  return 0;
}

