/*
 * This file is part of decoding
 * Copyright 2012  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#define _POSIX_C_SOURCE 1

#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <decoding/decoding.h>

#define RING_NAME fld
#define PRIME 67
#include <decoding/rings/gfp_word.c>
#include <decoding/algos.c>

void mtest_upol_horner(FILE *out,int n1,int n2) {
  int i;
  fld *f,x,y;
  f = fld_ll_upol_malloc_random(n2);
  fld_init(x);
  fld_init(y);
  for( i = n1 ; i <= n2 ; i++ ) {
    fld_random(x);
    fld_ll_upol_horner(y,f,i,x);
    fld_ll_print(out,"if e eq 0 then\n"
                     "  f  := P ! (%p);\n"
                     "  x  := fld ! (%e);\n"
                     "  y  := fld ! (%e);\n"
                     "  if not (Evaluate(f,x) eq y) then\n"
                     "    print \"Error\";\n"
                     "    e := 1;\n"
                     "  end if;\n"
                     "end if;\n\n",
                     f,i,x,y);
  }
  fld_ll_vec_free(f,n2);
  fld_clear(x);
  fld_clear(y);
}

void mtest_upol_mul(FILE *out,int n1,int n2) {
  int i,j,m;
  fld *f,*g,*fg;
  m = 2 * n2 - 1;
  f = fld_ll_upol_malloc_random(n2);
  g = fld_ll_upol_malloc_random(n2);
  fg = fld_ll_vec_malloc_init(m);
  for( i = n1 ; i <= n2 ; i++ ) {
    j = i / 3;
    fld_ll_upol_mul(fg,f,i,g,j);
    fld_ll_print(out,"if e eq 0 then\n"
                     "  f  := P ! (%p);\n"
                     "  g  := P ! (%p);\n"
                     "  fg := P ! (%p);\n"
                     "  if not (fg eq (f * g)) then\n"
                     "    print \"Error\";\n"
                     "    e := 1;\n"
                     "  end if;\n"
                     "end if;\n\n",
                     f,i,g,j,fg,i + j - 1);
  }
  fld_ll_vec_free(f,n2);
  fld_ll_vec_free(g,n2);
  fld_ll_vec_free(fg,m);
}

void mtest_upol_square(FILE *out,int n1,int n2) {
  int i,m;
  fld *f,*f2;
  m = 2 * n2 - 1;
  f = fld_ll_upol_malloc_random(n2);
  f2 = fld_ll_vec_malloc_init(m);
  for( i = n1 ; i <= n2 ; i++ ) {
    fld_ll_upol_square(f2,f,i);
    fld_ll_print(out,"if e eq 0 then\n"
                     "  f  := P ! (%p);\n"
                     "  f2 := P ! (%p);\n"
                     "  if not (f2 eq (f * f)) then\n"
                     "    print \"Error\";\n"
                     "    e := 1;\n"
                     "  end if;\n"
                     "end if;\n\n",
                     f,i,f2,2 * i - 1);
  }
  fld_ll_vec_free(f,n2);
  fld_ll_vec_free(f2,m);
}
void mtest_upol_eucldiv_naive(FILE *out,int n1,int n2) {
  int i,j,nr;
  fld *f,*g,*r,*q;
  f = fld_ll_vec_malloc_init(n2);
  g = fld_ll_vec_malloc_init(n2);
  r = fld_ll_vec_malloc_init(n2);
  q = fld_ll_vec_malloc_init(n2);
  for( i = n1 ; i <= n2 ; i++ ) {
    j = i / 3;
    fld_ll_upol_random(f,i);
    fld_ll_upol_random(g,j);
    fld_ll_upol_quorem_naive(r,&nr,q,f,i,g,j);
    fld_ll_print(out,"if e eq 0 then\n"
                     "  f := P ! (%p);\n"
                     "  g := P ! (%p);\n"
                     "  r := P ! (%p);\n"
                     "  q := P ! (%p);\n"
                     "  if not ( (q eq (f div g)) and\n"
                     "           (r eq (f mod g)) and\n"
                     "           (Degree(r) eq (%d - 1)) ) then\n"
                     "    print \"Error\";\n"
                     "    e := 1;\n"
                     "  end if;\n"
                     "end if;\n\n",
                     f,i,g,j,r,nr,q,i - j + 1,nr);
  }
  fld_ll_vec_free(f,n2);
  fld_ll_vec_free(g,n2);
  fld_ll_vec_free(r,n2);
  fld_ll_vec_free(q,n2);
}

void mtest_upol_eucldiv2_naive(FILE *out,int n1,int n2) {
  int i,j,nr;
  fld *f,*g,*r,*q;
  f = fld_ll_vec_malloc_init(n2);
  g = fld_ll_vec_malloc_init(n2);
  r = fld_ll_vec_malloc_init(n2);
  q = fld_ll_vec_malloc_init(n2);
  for( i = n1 ; i <= n2 ; i++ ) {
    j = i / 3;
    fld_ll_upol_random(f,i);
    fld_ll_upol_random(g,j);
    fld_ll_upol_quo_naive(q,f,i,g,j);
    fld_ll_upol_rem_naive(r,&nr,f,i,g,j);
    fld_ll_print(out,"if e eq 0 then\n"
                     "  f := P ! (%p);\n"
                     "  g := P ! (%p);\n"
                     "  r := P ! (%p);\n"
                     "  q := P ! (%p);\n"
                     "  if not ( (q eq (f div g)) and\n"
                     "           (r eq (f mod g)) and\n"
                     "           (Degree(r) eq (%d - 1)) ) then\n"
                     "    print \"Error\";\n"
                     "    e := 1;\n"
                     "  end if;\n"
                     "end if;\n\n",
                     f,i,g,j,r,nr,q,i - j + 1,nr);
  }
  fld_ll_vec_free(f,n2);
  fld_ll_vec_free(g,n2);
  fld_ll_vec_free(r,n2);
  fld_ll_vec_free(q,n2);
}

void mtest_upol_gcd_naive(FILE *out,int n1,int n2) {
  int i,j,ngcd;
  fld *f,*g,*gcd;
  f = fld_ll_vec_malloc_init(n2);
  g = fld_ll_vec_malloc_init(n2);
  gcd = fld_ll_vec_malloc_init(n2);
  for( i = n1 ; i <= n2 ; i++ ) {
    j = i / 3;
    fld_ll_upol_random(f,i);
    fld_ll_upol_random(g,j);
    fld_ll_upol_gcd_naive(gcd,&ngcd,f,i,g,j);
    fld_ll_upol_normalize(gcd,ngcd);
    fld_ll_print(out,"if e eq 0 then\n"
                     "  f   := P ! (%p);\n"
                     "  g   := P ! (%p);\n"
                     "  gcd := P ! (%p);\n"
                     "  if not (gcd eq Gcd(f,g)) then\n"
                     "    print \"Error\";\n"
                     "    e := 1;\n"
                     "  end if;\n"
                     "end if;\n\n",
                     f,i,g,j,gcd,ngcd);
  }
  fld_ll_vec_free(f,n2);
  fld_ll_vec_free(g,n2);
  fld_ll_vec_free(gcd,n2);
}

void mtest_upol_shift_naive(FILE *out,int n1,int n2) {
  int i;
  fld *f,*g,b;
  f = fld_ll_upol_malloc_random(n2);
  g = fld_ll_vec_malloc_init(n2);
  fld_init(b);
  for( i = n1 ; i <= n2 ; i++ ) {
    fld_random(b);
    fld_ll_upol_shift_naive(g,f,i,b);
    fld_ll_print(out,"if e eq 0 then\n"
                     "  f  := P ! (%p);\n"
                     "  b  := fld ! (%e);\n"
                     "  g  := P ! (%p);\n"
                     "  if not (Evaluate(f,X + b) eq g) then\n"
                     "    print \"Error\";\n"
                     "    e := 1;\n"
                     "  end if;\n"
                     "end if;\n\n",
                     f,i,b,g,i);
  }
  fld_ll_vec_free(f,n2);
  fld_ll_vec_free(g,n2);
  fld_clear(b);
}

void mtest_upol_shift_fast(FILE *out,int n1,int n2) {
  int i;
  fld *f,*g,b;
  f = fld_ll_upol_malloc_random(n2);
  g = fld_ll_vec_malloc_init(n2);
  fld_init(b);
  for( i = n1 ; i <= n2 ; i++ ) {
    fld_random(b);
    fld_ll_upol_shift_fast(g,f,i,b);
    fld_ll_print(out,"if e eq 0 then\n"
                     "  f  := P ! (%p);\n"
                     "  b  := fld ! (%e);\n"
                     "  g  := P ! (%p);\n"
                     "  if not (Evaluate(f,X + b) eq g) then\n"
                     "    print \"Error\";\n"
                     "    e := 1;\n"
                     "  end if;\n"
                     "end if;\n\n",
                     f,i,b,g,i);
  }
  fld_ll_vec_free(f,n2);
  fld_ll_vec_free(g,n2);
  fld_clear(b);
}

void mtest_upol_prod_affine(FILE *out,int n1,int n2) {
  int i;
  fld *f,*g,b;
  f = fld_ll_upol_malloc_random(n2 + 1);
  g = fld_ll_vec_malloc_init(n2 + 1);
  fld_init(b);
  for( i = n1 ; i <= n2 ; i++ ) {
    fld_ll_upol_prod_affine(g,f,i);
    fld_ll_print(out,"if e eq 0 then\n"
                     "  f  := %v;\n"
                     "  g  := P ! (%p);\n"
                     "  if not ((&*[(X + a) : a in f]) eq g) then\n"
                     "    print \"Error\";\n"
                     "    e := 1;\n"
                     "  end if;\n"
                     "end if;\n\n",
                     f,i,g,i + 1);
  }
  fld_ll_vec_free(f,n2 + 1);
  fld_ll_vec_free(g,n2 + 1);
  fld_clear(b);
}

void mtest_upol_prod_affine_neg(FILE *out,int n1,int n2) {
  int i;
  fld *f,*g,b;
  f = fld_ll_upol_malloc_random(n2 + 1);
  g = fld_ll_vec_malloc_init(n2 + 1);
  fld_init(b);
  for( i = n1 ; i <= n2 ; i++ ) {
    fld_ll_upol_prod_affine_neg(g,f,i);
    fld_ll_print(out,"if e eq 0 then\n"
                     "  f  := %v;\n"
                     "  g  := P ! (%p);\n"
                     "  if not ((&*[(X - a) : a in f]) eq g) then\n"
                     "    print \"Error\";\n"
                     "    e := 1;\n"
                     "  end if;\n"
                     "end if;\n\n",
                     f,i,g,i + 1);
  }
  fld_ll_vec_free(f,n2 + 1);
  fld_ll_vec_free(g,n2 + 1);
  fld_clear(b);
}

#define D1 5
#define D2 100

int main(void) {
  srand(time(0));
  fld_ring_init();
  fld_ll_print(stdout,"clear;\n"
                      "%s;\n"
                      "P<X> := PolynomialRing(k);\n"
                      "e := 0;\n\n",fld_magma_decl);
  mtest_upol_horner(stdout,D1,D2);
  mtest_upol_mul(stdout,D1,D2);
  mtest_upol_square(stdout,D1,D2);
  mtest_upol_eucldiv_naive(stdout,D1,D2);
  mtest_upol_eucldiv2_naive(stdout,D1,D2);
  mtest_upol_gcd_naive(stdout,D1,D2);
  mtest_upol_shift_naive(stdout,D1,D2);
  mtest_upol_shift_fast(stdout,D1,D2);
  mtest_upol_prod_affine(stdout,D1,D2);
  mtest_upol_prod_affine_neg(stdout,D1,D2);
  fld_ll_print(stdout,"if e eq 0 then\n"
                      "  print \"ALL OK\";\n"
                      "end if;\n\n");
  return 0;
}

