/*
 * This file is part of decoding
 * Copyright 2012  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#define _POSIX_C_SOURCE 1

#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <decoding/decoding.h>

#define RING_NAME gf2n
#define DEGREE 2
#include <decoding/rings/mpfq_gf2n_wrapper.c>
#include <decoding/algos.c>

void mtest_upol_horner(FILE *out,int n1,int n2) {
  int i;
  gf2n *f,x,y;
  f = gf2n_ll_upol_malloc_random(n2);
  gf2n_init(x);
  gf2n_init(y);
  for( i = n1 ; i <= n2 ; i++ ) {
    gf2n_random(x);
    gf2n_ll_upol_horner(y,f,i,x);
    gf2n_ll_print(out,"if e eq 0 then\n"
                      "  f  := P ! (%p);\n"
                      "  x  := k ! (%e);\n"
                      "  y  := k ! (%e);\n"
                      "  if not (Evaluate(f,x) eq y) then\n"
                      "    print \"Error\";\n"
                      "    e := 1;\n"
                      "  end if;\n"
                      "end if;\n\n",
                      f,i,x,y);
  }
  gf2n_ll_vec_free(f,n2);
  gf2n_clear(x);
  gf2n_clear(y);
}

void mtest_upol_mul(FILE *out,int n1,int n2) {
  int i,j,m;
  gf2n *f,*g,*fg;
  m = 2 * n2 - 1;
  f = gf2n_ll_upol_malloc_random(n2);
  g = gf2n_ll_upol_malloc_random(n2);
  fg = gf2n_ll_vec_malloc_init(m);
  for( i = n1 ; i <= n2 ; i++ ) {
    j = i / 3;
    gf2n_ll_upol_mul(fg,f,i,g,j);
    gf2n_ll_print(out,"if e eq 0 then\n"
                 "  f  := P ! (%p);\n"
                 "  g  := P ! (%p);\n"
                 "  fg := P ! (%p);\n"
                 "  if not (fg eq (f * g)) then\n"
                 "    print \"Error\";\n"
                 "    e := 1;\n"
                 "  end if;\n"
                 "end if;\n\n",
                 f,i,g,j,fg,i + j - 1);
  }
  gf2n_ll_vec_free(f,n2);
  gf2n_ll_vec_free(g,n2);
  gf2n_ll_vec_free(fg,m);
}

void mtest_upol_square(FILE *out,int n1,int n2) {
  int i,m;
  gf2n *f,*f2;
  m = 2 * n2 - 1;
  f = gf2n_ll_upol_malloc_random(n2);
  f2 = gf2n_ll_vec_malloc_init(m);
  for( i = n1 ; i <= n2 ; i++ ) {
    gf2n_ll_upol_square(f2,f,i);
    gf2n_ll_print(out,"if e eq 0 then\n"
                 "  f  := P ! (%p);\n"
                 "  f2 := P ! (%p);\n"
                 "  if not (f2 eq (f * f)) then\n"
                 "    print \"Error\";\n"
                 "    e := 1;\n"
                 "  end if;\n"
                 "end if;\n\n",
                 f,i,f2,2 * i - 1);
  }
  gf2n_ll_vec_free(f,n2);
  gf2n_ll_vec_free(f2,m);
}
void mtest_upol_eucldiv_naive(FILE *out,int n1,int n2) {
  int i,j,nr;
  gf2n *f,*g,*r,*q;
  f = gf2n_ll_vec_malloc_init(n2);
  g = gf2n_ll_vec_malloc_init(n2);
  r = gf2n_ll_vec_malloc_init(n2);
  q = gf2n_ll_vec_malloc_init(n2);
  for( i = n1 ; i <= n2 ; i++ ) {
    j = i / 3;
    gf2n_ll_upol_random(f,i);
    gf2n_ll_upol_random(g,j);
    gf2n_ll_upol_quorem_naive(r,&nr,q,f,i,g,j);
    gf2n_ll_print(out,"if e eq 0 then\n"
                 "  f := P ! (%p);\n"
                 "  g := P ! (%p);\n"
                 "  r := P ! (%p);\n"
                 "  q := P ! (%p);\n"
                 "  if not ( (q eq (f div g)) and\n"
                 "           (r eq (f mod g)) and\n"
                 "           (Degree(r) eq (%d - 1)) ) then\n"
                 "    print \"Error\";\n"
                 "    e := 1;\n"
                 "  end if;\n"
                 "end if;\n\n",
                 f,i,g,j,r,nr,q,i - j + 1,nr);
  }
  gf2n_ll_vec_free(f,n2);
  gf2n_ll_vec_free(g,n2);
  gf2n_ll_vec_free(r,n2);
  gf2n_ll_vec_free(q,n2);
}

void mtest_upol_eucldiv2_naive(FILE *out,int n1,int n2) {
  int i,j,nr;
  gf2n *f,*g,*r,*q;
  f = gf2n_ll_vec_malloc_init(n2);
  g = gf2n_ll_vec_malloc_init(n2);
  r = gf2n_ll_vec_malloc_init(n2);
  q = gf2n_ll_vec_malloc_init(n2);
  for( i = n1 ; i <= n2 ; i++ ) {
    j = i / 3;
    gf2n_ll_upol_random(f,i);
    gf2n_ll_upol_random(g,j);
    gf2n_ll_upol_quo_naive(q,f,i,g,j);
    gf2n_ll_upol_rem_naive(r,&nr,f,i,g,j);
    gf2n_ll_print(out,"if e eq 0 then\n"
                 "  f := P ! (%p);\n"
                 "  g := P ! (%p);\n"
                 "  r := P ! (%p);\n"
                 "  q := P ! (%p);\n"
                 "  if not ( (q eq (f div g)) and\n"
                 "           (r eq (f mod g)) and\n"
                 "           (Degree(r) eq (%d - 1)) ) then\n"
                 "    print \"Error\";\n"
                 "    e := 1;\n"
                 "  end if;\n"
                 "end if;\n\n",
                 f,i,g,j,r,nr,q,i - j + 1,nr);
  }
  gf2n_ll_vec_free(f,n2);
  gf2n_ll_vec_free(g,n2);
  gf2n_ll_vec_free(r,n2);
  gf2n_ll_vec_free(q,n2);
}

void mtest_upol_gcd_naive(FILE *out,int n1,int n2) {
  int i,j,ngcd;
  gf2n *f,*g,*gcd;
  f = gf2n_ll_vec_malloc_init(n2);
  g = gf2n_ll_vec_malloc_init(n2);
  gcd = gf2n_ll_vec_malloc_init(n2);
  for( i = n1 ; i <= n2 ; i++ ) {
    j = i / 3;
    gf2n_ll_upol_random(f,i);
    gf2n_ll_upol_random(g,j);
    gf2n_ll_upol_gcd_naive(gcd,&ngcd,f,i,g,j);
    gf2n_ll_upol_normalize(gcd,ngcd);
    gf2n_ll_print(out,"if e eq 0 then\n"
                 "  f   := P ! (%p);\n"
                 "  g   := P ! (%p);\n"
                 "  gcd := P ! (%p);\n"
                 "  if not (gcd eq Gcd(f,g)) then\n"
                 "    print \"Error\";\n"
                 "    e := 1;\n"
                 "  end if;\n"
                 "end if;\n\n",
                 f,i,g,j,gcd,ngcd);
  }
  gf2n_ll_vec_free(f,n2);
  gf2n_ll_vec_free(g,n2);
  gf2n_ll_vec_free(gcd,n2);
}

void mtest_upol_shift_naive(FILE *out,int n1,int n2) {
  int i;
  gf2n *f,*g,b;
  f = gf2n_ll_upol_malloc_random(n2);
  g = gf2n_ll_vec_malloc_init(n2);
  gf2n_init(b);
  for( i = n1 ; i <= n2 ; i++ ) {
    gf2n_random(b);
    gf2n_ll_upol_shift_naive(g,f,i,b);
    gf2n_ll_print(out,"if e eq 0 then\n"
                 "  f  := P ! (%p);\n"
                 "  b  := k ! (%e);\n"
                 "  g  := P ! (%p);\n"
                 "  if not (Evaluate(f,X + b) eq g) then\n"
                 "    print \"Error\";\n"
                 "    e := 1;\n"
                 "  end if;\n"
                 "end if;\n\n",
                 f,i,b,g,i);
  }
  gf2n_ll_vec_free(f,n2);
  gf2n_ll_vec_free(g,n2);
  gf2n_clear(b);
}

void mtest_upol_shift_fast(FILE *out,int n1,int n2) {
  int i;
  gf2n *f,*g,b;
  f = gf2n_ll_upol_malloc_random(n2);
  g = gf2n_ll_vec_malloc_init(n2);
  gf2n_init(b);
  for( i = n1 ; i <= n2 ; i++ ) {
    gf2n_random(b);
    gf2n_ll_upol_shift_fast(g,f,i,b);
    gf2n_ll_print(out,"if e eq 0 then\n"
                 "  f  := P ! (%p);\n"
                 "  b  := k ! (%e);\n"
                 "  g  := P ! (%p);\n"
                 "  if not (Evaluate(f,X + b) eq g) then\n"
                 "    print \"Error\";\n"
                 "    e := 1;\n"
                 "  end if;\n"
                 "end if;\n\n",
                 f,i,b,g,i);
  }
  gf2n_ll_vec_free(f,n2);
  gf2n_ll_vec_free(g,n2);
  gf2n_clear(b);
}

void mtest_upol_shift_car2(FILE *out,int n1,int n2) {
  int i;
  gf2n *f,*g,b;
  f = gf2n_ll_upol_malloc_random(n2);
  g = gf2n_ll_vec_malloc_init(n2);
  gf2n_init(b);
  for( i = n1 ; i <= n2 ; i++ ) {
    gf2n_random(b);
    gf2n_ll_upol_shift_car2(g,f,i,b);
    gf2n_ll_print(out,"if e eq 0 then\n"
                 "  f  := P ! (%p);\n"
                 "  b  := k ! (%e);\n"
                 "  g  := P ! (%p);\n"
                 "  if not (Evaluate(f,X + b) eq g) then\n"
                 "    print \"Error\";\n"
                 "    e := 1;\n"
                 "  end if;\n"
                 "end if;\n\n",
                 f,i,b,g,i);
  }
  gf2n_ll_vec_free(f,n2);
  gf2n_ll_vec_free(g,n2);
  gf2n_clear(b);
}

void mtest_upol_shift_car2_inplace(FILE *out,int n1,int n2) {
  int i;
  gf2n *f,*g,b;
  f = gf2n_ll_upol_malloc_random(n2);
  g = gf2n_ll_vec_malloc_init(n2);
  gf2n_init(b);
  for( i = n1 ; i <= n2 ; i++ ) {
    gf2n_random(b);
    gf2n_ll_vec_copy(g,f,i);
    gf2n_ll_upol_shift_car2(g,g,i,b);
    gf2n_ll_print(out,"if e eq 0 then\n"
                 "  f  := P ! (%p);\n"
                 "  b  := k ! (%e);\n"
                 "  g  := P ! (%p);\n"
                 "  if not (Evaluate(f,X + b) eq g) then\n"
                 "    print \"Error\";\n"
                 "    e := 1;\n"
                 "  end if;\n"
                 "end if;\n\n",
                 f,i,b,g,i);
  }
  gf2n_ll_vec_free(f,n2);
  gf2n_ll_vec_free(g,n2);
  gf2n_clear(b);
}

#define D1 5
#define D2 10

int main(void) {
  ll_random_seed(time(0));
  gf2n_ring_init();
  gf2n_ll_print(stdout,"clear;\n"
                  "%s;\n"
                  "P<X> := PolynomialRing(k);\n"
                  "e := 0;\n\n",gf2n_magma_decl);
  mtest_upol_horner(stdout,D1,D2);
  /*
  mtest_upol_mul(stdout,D1,D2);
  mtest_upol_square(stdout,D1,D2);
  mtest_upol_eucldiv_naive(stdout,D1,D2);
  mtest_upol_eucldiv2_naive(stdout,D1,D2);
  mtest_upol_gcd_naive(stdout,D1,D2);
  mtest_upol_shift_naive(stdout,D1,D2);
  mtest_upol_shift_fast(stdout,D1,D2);
  mtest_upol_shift_car2(stdout,D1,D2);
  mtest_upol_shift_car2_inplace(stdout,D1,D2);
  */
  gf2n_ll_print(stdout,"if e eq 0 then\n"
                  "  print \"ALL OK\";\n"
                  "end if;\n\n");
  return 0;
}

