/*
 * This file is part of decoding
 * Copyright 2012  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#define _POSIX_C_SOURCE 1

#include <stdio.h>
#include <decoding/decoding.h>

#define RING_NAME fld
#include <decoding/rings/gfp_word_dynamic.c>
#include <decoding/algos.c>

void test_inv(int n) {
  int i;
  fld a,b,c,one;
  fld_init(a);
  fld_init(b);
  fld_init(c);
  fld_init(one);
  fld_castint(one,1);
  for( i = 0 ; i < n ; i++ ) {
    fld_random(a);
    if ( fld_iszero(a) ) fld_castint(a,1);
    if ( fld_inv(b,a) ) {
      fld_mul(c,a,b);
      if ( fld_cmp(c,one) ) {
        fld_ll_print(stderr,"Error (%e)^-1 != %e.\n",a,b);
        exit(-1);
      }
    }
  }
  fld_clear(a);
  fld_clear(b);
  fld_clear(c);
  fld_clear(one);
  fputs("test_inv: OK\n",stdout);
}

void test_add(int n) {
  int i;
  fld a,b,c;
  fld_init(a);
  fld_init(b);
  fld_init(c);
  for( i = 0 ; i < n ; i++ ) {
    fld_random(b);
    fld_random(c);
    fld_add(a,b,c);
    if ( a[0] != (b[0] + c[0]) % fld_characteristic ) {
      fld_ll_print(stderr,"[%d] error: %e != %e + %e.\n",i,a,b,c);
      exit(-1);
    }
  }
  fld_clear(a);
  fld_clear(b);
  fld_clear(c);
  fputs("test_add: OK\n",stdout);
}

void test_mul(int n) {
  int i;
  fld a,b,c;
  fld_init(a);
  fld_init(b);
  fld_init(c);
  for( i = 0 ; i < n ; i++ ) {
    fld_random(b);
    fld_random(c);
    fld_mul(a,b,c);
    if ( a[0] != ((b[0] * c[0]) % fld_characteristic) ) {
      fld_ll_print(stderr,"[%d] error: %e != %e * %e.\n",i,a,b,c);
      exit(-1);
    }
  }
  fld_clear(a);
  fld_clear(b);
  fld_clear(c);
  fputs("test_mul: OK\n",stdout);
}

void test_ll_upol_mul(int n) {
  int i;
  fld *a,*b,*c,*d;
  a = fld_ll_vec_malloc_init(n);
  b = fld_ll_vec_malloc_init(n);
  c = fld_ll_vec_malloc_init(2 * n);
  d = fld_ll_vec_malloc_init(2 * n);
  for( i = 2 ; i < n ; i++ ) {
    fld_ll_vec_random(a,i);
    fld_ll_vec_random(b,i);
    fld_ll_upol_mul(c,a,i,b,i);
    fld_ll_upol_mul_naive(d,a,i,b,i);
    if ( fld_ll_vec_cmp(c,d,2 * i - 1) ) {
      fld_ll_print(stderr,"[%d] error:\n"
                          "a      = %p\n"
                          "b      = %p\n"
                          "a * b != %p\n",i,a,i,b,i,c,2 * i - 1);
      exit(-1);
    }
  }
  fld_ll_vec_free(a,n);
  fld_ll_vec_free(b,n);
  fld_ll_vec_free(c,2 * n);
  fld_ll_vec_free(d,2 * n);
  fputs("test_ll_upol_mul: OK\n",stdout);
}

void test_ll_dbpol_mul(int n) {
  int i,m;
  fld *a,*b,*c,*d;
  a = fld_ll_vec_malloc_init(n * n);
  b = fld_ll_vec_malloc_init(n * n);
  c = fld_ll_vec_malloc_init(4 * n * n);
  d = fld_ll_vec_malloc_init(4 * n * n);
  for( i = 2 ; i < n ; i++ ) {
    m = 2 * i - 1;
    fld_ll_vec_random(a,i * i);
    fld_ll_vec_random(b,i * i);
    fld_ll_dbpol_mul(c,a,i,i,b,i,i);
    fld_ll_dbpol_mul_naive(d,a,i,i,b,i,i);
    if ( fld_ll_vec_cmp(c,d,m * m) ) {
      fld_ll_print(stderr,"[%d] error:\n"
                          "a      = %q\n"
                          "b      = %q\n"
                          "a * b != %q\n",i,a,i,i,b,i,i,c,m,m);
      exit(-1);
    }
  }
  fld_ll_vec_free(a,n * n);
  fld_ll_vec_free(b,n * n);
  fld_ll_vec_free(c,4 * n * n);
  fld_ll_vec_free(d,4 * n * n);
  fputs("test_ll_dbpol_mul: OK\n",stdout);
}

int main(void) {
  fld_ring_init(101);
  print_compile_flags(stdout);
  test_inv(100000);
  test_add(100000);
  test_mul(100000);
  test_ll_upol_mul(100);
  test_ll_dbpol_mul(100);
  return 0;
}

